#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#


# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_builtin_eval.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

#
# Test whether the "eval" special shell builtin properly restores
# the I/O redirection
#
# This was reported as CR #6988358 ("ksh93 mishandles GNU ld genscripts.sh script"):
# ------------ snip ------------
# When trying to build GNU ld from the GNU binutils CVS repository on
# a self-built snv_147, the ld/genscripts.sh failed, which doesn't
# happen with the former
# /bin/sh (i.e. /usr/has/bin/sh).
# Frequency
#      Always
# Regression
#      solaris_10
# Steps to Reproduce
#      Consider the following script:
#      % cat genscripts.sh
#      #!/bin/sh
#      TMP=tmp
#      (
#         . source.sh
#      ) | cat > out
#      rm -f ${TMP}
#      % cat source.sh
#      eval "cat > $TMP" <<EOF
#      EOF
#      sed -e '1d' $TMP
#      sed -e '2d' $TMP
#      % chmod +x genscripts.sh
#      % ./genscripts.sh
#      % sed: can't read tmp: No such file or directory
# Expected Result
#      The scripts runs without errors, as it does with both 
#      /usr/has/bin/sh and bash.
# Actual Result
#      The script errors out as above, obviously removing ${TMP} too early
#      Unfortunately, the error vanishes if running genscripts.sh with sh -x, 
#      unlike the full genscripts.sh from
#      http://sourceware.org/cgi-bin/cvsweb.cgi/src/ld/genscripts.sh?rev=1.31&content-type=text/x-cvsweb-markup&cvsroot=src
#      where the early removal can clearly be seen when running with sh -x.
# ------------ snip ------------
#
function test_eval_restore_io_redirection
{
	float sec
	typeset testname
	typeset currtest # current test line
	typeset scriptname # name of script to be sourced
	typeset sourcetype # type of "source" being used
	typeset catcmd # "cat" being used (builtin or external)
	integer ilsi
	compound out=( typeset stdout stderr ; integer res )
	typeset scout # shell compiler output
	
	typeset -r -a inlinescripts=(
		' eval " cat  > /dev/null  < /dev/null"  ; sleep 4\n'
		' eval "(cat  > /dev/null  < /dev/null)" ; sleep 4\n'
		' eval "(cat) > /dev/null  < /dev/null"  ; sleep 4\n'
		'(eval "cat   > /dev/null  < /dev/null") ; sleep 4\n'
		# same as above but uses "ulimit -c 0" to force the
		# shell to use |fork()| for the subshell
		' eval "(ulimit -c 0 ; cat  > /dev/null  < /dev/null)" ; sleep 4\n'
		' eval "(ulimit -c 0 ; cat) > /dev/null  < /dev/null"  ; sleep 4\n'
		'(ulimit -c 0 ; eval "cat   > /dev/null  < /dev/null") ; sleep 4\n'
	)

	for (( ilsi=0 ; ilsi < ${#inlinescripts[@]} ; ilsi++ )) ; do
		# create inline script (and provide a compiled version, too)	
		printf "${inlinescripts[ilsi]}" >'foo.sh'
		scout="$( ${SHCOMP} 'foo.sh' 'foo.shbin' 2>&1 )" || err_exit "shcomp failed with exit code $?"
		[[ "${scout}" == '' ]] || err_exit "Unexpected shcomp compiler output: ${ printf '%q\n' "${scout}" ; }"

		for scriptname in \
			'./foo.sh' \
			'./foo.shbin' ; do	
			# iterate through tests
			for catcmd in \
				'cat' \
				'/usr/bin/cat' ; do
				for sourcetype in \
					'.' \
					'source' ; do
					testname="test_${0}/${catcmd}/${scriptname}/${sourcetype}"
				
					currtest='builtin cat ; @@sourcetype@@ @@scriptname@@ | @@catcmd@@ > /dev/null ; true'
					currtest="${currtest/@@sourcetype@@/${sourcetype}}"
					currtest="${currtest/@@scriptname@@/${scriptname}}"
					currtest="${currtest/@@catcmd@@/${catcmd}}"
		
					(( sec=SECONDS ))	
	        			out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${currtest}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
					(( (SECONDS-sec) < 2. ))  && err_exit "${testname}: inline script does not restore output redirection with eval"

					[[ "${out.stdout}" == '' ]] || err_exit "${testname}: Expected empty stdout, got ${ printf '%q\n' "${out.stdout}" ; }"
					[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
					(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
				done
			done
		done

		rm 'foo.sh' 'foo.shbin'
	done
	
	return 0
}

# test scope of "var=<value> eval ..."
function test_variable_scope1
{
	typeset testname
	compound out=( typeset stdout stderr ; integer res )
	compound -r -a tests=(
		( cmd='avar=2 ; avar=1 eval "(print \$avar ; unset avar)" ; printf "%s\n" "$avar"' expected_output=$'1\n1' )
		( cmd='avar=2 ; avar=1 eval " print \$avar ; unset avar " ; printf "%s\n" "$avar"' expected_output=$'1' )
	)
	integer ti # test index
	
	for (( ti=0 ; ti < ${#tests[@]} ; ti++ )) ; do
		nameref test=tests[ti]
		
		testname="${0}/${ti}"
	
		out.stderr="${ { out.stdout="${ ${SHELL} -c "${test.cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
	
		[[ "${out.stdout}" == "${test.expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${test.expected_output}" ; }, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == ''                        ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}



# run tests
test_eval_restore_io_redirection
test_variable_scope1


cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
