#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# The test module check whether sedid scripts work as expected.
#
# This was reported as CR #6934836 ("set-uid script with -p in magic
# number gets Exec format error"):
# ------------ snip ------------
# The following script, when run as a set-uid script, fails to execute,
# reporting an "Exec format error" instead:
# 
# 
# $ cat ok.ksh
# #!/bin/ksh -p
# 
# print "okay"
# $
# 
# This is odd because ksh -p is documented as doing nothing more than
# suppressing the processing of the user's $HOME/.profile and using
# /etc/suid_profile instead of the ENV variable.  Nothing in available
# ksh93 documentation suggests that -p is incompatible with set-uid
# scripts.
# 
# Does not seem to have anything to do with the "-p" option 
# specifically.  The script fails in  exactly the same way if we 
# use any options, "-x", "-v", etc.
# 
# The difference between the way a suid script is executed and a 
# regular one is that in case of a suid script, the script to be
# executed is passed as a /dev/fd parameter to close the "replace
# script between exec and open by shell" hole 
# (usr/src/uts/common/exec/shbin/shbin.c). This, with ksh93, 
# seems to have a problem *only* if there are options specified 
# on the command line.
# 
# The program below demonstrates this:
# 
# $ cat testexec.c
# #include <stdio.h>
# #include <fcntl.h>
# #include <unistd.h>
# 
# int
# main()
# {
#         int fd = -1;
#         char devfd[32];
#         char *script = "/tmp/ok.ksh";
# 
#         fd = open(script, O_RDONLY);
#         sprintf(devfd, "/dev/fd/%d", fd);
#         execl("/usr/bin/sparcv9/ksh93", "ksh", "-v", devfd, NULL);
#         //execl("/usr/bin/sparcv9/ksh93", "ksh", "-v", script, NULL);
#         //execl("/usr/bin/bash", "bash", "-v", devfd, NULL);
#         //execl("/usr/bin/sparcv9/ksh93", "ksh", devfd, NULL);
#         perror("exec failed");
# }
# $ ./testexec
# /usr/bin/ksh: /usr/bin/ksh: cannot execute [Exec format error]
# 
# This fails irrespective of whether the script is suid or not.
# And the other commented execl() lines in the source (using the 
# script directly instead of the /dev/fd path, bash with command
# line options and the devfd path, or ksh93 without command line
# options) all work perfectly.
#
# A few interesting twists:
# 
# $ cp testexec kshexec
# $ ./kshexec
# okay
# 
# (Renaming the script to something starting with 'ksh' makes it
# start working.)
# 
# $ SHELL=/bin/ksh ./testexec
# ksh: ksh: cannot execute [Exec format error]
# $ SHELL=/bin/bash ./testexec
# /usr/bin/ksh: /usr/bin/ksh: cannot execute binary file
# 
# (Though we are invoking ksh93 directly, it does an exec of itself
# when we use /dev/fd/XX scripts where it ends up using the SHELL
# environment variable.)
# The culprit seems to be the code below:
# 
# <lib/libshell/common/sh/init.c>
# 1216   shp->st.dolv=argv+(argc-1)-shp->st.dolc;
# 1217   shp->st.dolv[0] = argv[0];
# 
# Here, we are overwriting one of the arguments of argv (because
# shp->st.dolv indexes into the argv vector).
# 
# In this particular case, argv which originally looked like this:
# 
#     ksh, -v, /dev/fd/3
# 
# ends up looking like this:
# 
#     ksh, ksh, /dev/fd/3
# 
# We then pass the mangled argv to execv():
# 
# <lib/libshell/common/sh/main.c>
# 
# 298   /*  exec to change $0 for ps  */
# 299   execv(pathshell(),av);
# 
# As a consequence, ksh tries to load the ksh binary as a shell 
# script and fails with an "Exec format" error.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="${ mktemp -t -d 'test_sun_solaris_cr_6934836_set-uid_script_with_-p_in_magic_number_gets_exec_format_error.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }


function compile_wrapper
{
	typeset PATH='/usr/xpg6/bin:/usr/xpg4/bin:/usr/ccs/bin:/usr/bin:/opt/SUNWspro/bin:/usr/gnu/bin'
	typeset out

	# our test wrapper executable
cat >"testwrap1.c" <<EOF
	#include <stdlib.h>
	#include <stdio.h>
	#include <fcntl.h>
	#include <unistd.h>
	
	int
	main(int ac, char *av[])
	{
		int fd = -1;
		char devfd[12+1+8+1]; /* sizeof(str(INT_MAX))+sign=strlen("/dev/fd/")+1 */
		const char *shell_name;
		const char *script_name;
		
		if (ac != 3)
		{
			(void)fprintf(stderr, "Not enough arguments, ac=%d\n", ac);
			return EXIT_FAILURE;
		}
		
		shell_name  = av[1];
		script_name = av[2];
	
		fd = open(script_name, O_RDONLY);
		if (fd == -1)
		{
			perror("Cannot open script");
			return EXIT_FAILURE;
		}
		(void)sprintf(devfd, "/dev/fd/%d", fd);
		(void)execl(shell_name, "ksh", "-v", devfd, NULL);
	
		perror("exec failed");
		return EXIT_FAILURE;
	}
EOF

	out="$(
		[[ "${!LD_*}" != '' ]] && unset ${!LD_*}
		cc -g -o testwrap1 testwrap1.c 2>&1
		)" || err_exit "Compiler failed with exit code $?"
	[[ "${out}" == '' ]] || err_exit "Unexpected compiler output: ${ printf '%q\n' "${out}" ; }"
	
	rm -f 'testwrap1.c'

	return 0
}

# test plain script
function test1_plain
{	
	typeset out

	# our test script
	print $'print "hello fish world\\n"\n' > 'testscript1.sh'

	# plain script
	out="$( ./testwrap1 "${SHELL}" 'testscript1.sh' 2>&1 )" || err_exit "Test script failed with exit code $?"
	[[ ! -f "core" ]] || err_exit "core dump found: ${ printf '%q\n' "$(pstack 'core')" ; }"
	[[ "${out}" == 'hello fish world' ]] || err_exit "Output should match 'hello fish world', got ${ printf '%q\n' "${out}" ; }"

	rm -f 'testscript1.sh'

	return 0
}

# test compiled script
function test1_compiled
{
	typeset out

	# our test script
	print $'print "hello fish world\\n"\n' > 'testscript1.sh'

	out="$( ${SHCOMP} 'testscript1.sh' 'testscript1.shbin' 2>&1 )" || err_exit "shcomp failed with exit code $?"
	[[ "${out}" == '' ]] || err_exit "Unexpected shcomp compiler output: ${ printf '%q\n' "${out}" ; }"
	
	out="$( ./testwrap1 "${SHELL}" "testscript1.shbin" 2>&1 )" || err_exit "Test script failed with exit code $?"
	[[ ! -f 'core' ]] || err_exit "core dump found: ${ printf '%q\n' "$(pstack 'core')" ; }"
	[[ "${out}" == 'hello fish world' ]] || err_exit "Output should match 'hello fish world', got ${ printf '%q\n' "${out}" ; }"
	
	rm -f 'testscript1.sh' 'testscript1.shbin'
	
	return 0
}

# run tests
compile_wrapper
test1_plain
test1_compiled

rm -f 'testwrap1'


# cleanup
cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."


# tests done
exit $((Errors<125?Errors:125))
