#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# The test module check whether mkdir -m 1777 correctly sets the "sticky"
# bit ('t') on directories.
#
# This was reported as CR #6941538 ("ksh93 mkdir -m 1777 fails to set sticky bit"):
# ------------ snip ------------
# # /bin/ksh93
# # type mkdir
# mkdir is a shell builtin version of /usr/bin/mkdir
# # mkdir -m 1777 /tmp/builtin
# # /usr/bin/mkdir -m 1777 /tmp/command
# # ls -ld /tmp/builtin /tmp/command
# drwxrwxrwx   2 root     root          69 Apr  7 12:19 /tmp/builtin
# drwxrwxrwt   2 root     root          69 Apr  7 12:19 /tmp/command
# Frequency
#    Always
# Regression
#    no
# Steps to Reproduce
#    See above.
# Expected Result
#    Sticky bit in -m argument is honored.
# Actual Result
#    Sticky bit is simply ignored.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="$(mktemp -t -d 'test_sun_solaris_cr_6941538_ksh93_mkdir_-m_1777_fails_to_set_sticky_bit.XXXXXXXX')" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }


# cr6941538/test 1: Sticky bit test
function test_cr6941538_1
{
	typeset out
	typeset mode
	typeset mkdircmd="$1"

	# test 1: test mode
	out="${ ${mkdircmd} -m 1777 'test1dir' 2>&1 ; }" || err_exit "${0}/${mkdircmd}: mkdir failed with exit code $?"
	[[ "${out}" == '' ]] || err_exit "${0}/${mkdircmd}: Output should be empty, got ${ printf '%q\n' "${out}" ; }" 
	ls -ld 'test1dir'  | read mode dummy ; [[ "${mode}" == d*t ]] || err_exit "${0}/builtin:  Expected mode match d*t, got '${mode}'"

	rmdir "test1dir"

	return 0
}

# cr6941538/test 2: Same as test1 but uses -p and a non-existing parent dir
function test_cr6941538_2
{
	typeset out
	typeset mode
	typeset mkdircmd="$1"

	out="${ ${mkdircmd} -p -m 1777 'p1/test2dir' 2>&1 ; }" || err_exit "${0}/${mkdircmd}: mkdir failed with exit code $?"
	[[ "${out}" == '' ]] || err_exit "${0}/${mkdircmd}: Output should be empty, got ${ printf '%q\n' "${out}" ; }" 
	ls -ld 'p1/test2dir' | read mode dummy ; [[ "${mode}" == d*t ]] || err_exit "${0}/${mkdircmd}: Expected mode match d*t, got '${mode}'"
	ls -ld 'p1'          | read mode dummy ; [[ "${mode}" == d*x ]] || err_exit "${0}/${mkdircmd}: Expected parent mode match d*t, got '${mode}'"

	rmdir 'p1/test2dir'
	rmdir 'p1'

	return 0
}


# enable mkdir builtin
builtin mkdir || err_exit "mkdir builtin not found, exit code $?"
builtin rmdir || err_exit "rmdir builtin not found, exit code $?"

# run tests
for cmd in 'mkdir' '/usr/gnu/bin/mkdir' '/usr/bin/mkdir' ; do
	test_cr6941538_1 "${cmd}"
	test_cr6941538_2 "${cmd}"
done


# cleanup
cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."


# tests done
exit $((Errors<125?Errors:125))
