#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# The test module verifies whether print -p can reach a coprocess
# when a command substitution is between proc |& and print -p
#
# This was reported as CR #6942280 ("ksh93 failure to write to process at other end of pipe"):
# ------------ snip ------------
# A script has two functions, one of which creates a two-way pipeline
# with '|&' and the other writes to the process at the other end of 
# the pipe using 'print -p'.  It is found that the writer fails with
#     "print: no query process"
# on ksh 93t 2008-11-04 and later.  The same failure is NOT seen on
# ksh 93t 2008-07-25 or with ksh88.  A tarball is attached containing
# the scripts which demonstrate the problem.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir
typeset out

# create temporary test directory
ocwd="$PWD"
tmpdir="${ mktemp -t -d 'test_sun_solaris_cr_6942280_print_p_cannot_find_coprocess.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }


# test1: Standalone version of original testcase
function test1
{
	compound out=( typeset stdout stderr ; integer res )
	typeset testscript

cat >'test1script.sh' <<EOF
	[[ ! -z \${TET_DEBUG} ]] && set -x

	readonly me=\${0##*/}
	readonly mydir=\${0%%/\$me}

	function tconfig
	{
		tet_kshapid_write "IN CONFIG ROUTINE"

		tet_kshapid_write "User activity snapshot on \$(hostname) at \$(date)"
		tet_kshapid_write "  "
		tet_kshapid_write "\$(who)"
	}

	function trun
	{
		tet_kshapid_write "IN RUN ROUTINE"

		tet_kshapid_write "User activity snapshot on \$(hostname) at \$(date)"
		tet_kshapid_write "  "
		tet_kshapid_write "\$(who)"
	}

	NUMBER_OF_TESTS=2
	set -A TET_TESTNAMES "tconfig" "trun"

	[[ ! -z \${TET_DEBUG} ]] && set -x

	typeset TET_ALARM_FLAG=no
	typeset -i TET_ALARM_PID=0
	typeset TET_API_VERSION=
	typeset -i TET_CONTEXT=0
	typeset -i TET_KSHAPID_PID=0
	typeset TET_KSHAPID_REQUEST=
	typeset TET_KSHAPID_USABLE=no
	typeset TET_MERROR_INPROGRESS=no
	typeset -r TET_PNAME=\$0
	typeset -i TET_REPLY_ARGC=0
	typeset TET_REPLY_ARGV
	typeset TET_REPLY_DATA_LINE
	typeset TET_TMPFILES=
	typeset -i TET_USYNC_NSYS=0
	typeset TET_USYNC_SYNCSTAT


	function tet_kshapid_startup
	{
		[[ ! -z \${TET_DEBUG} ]] && set -x
		typeset tet_l14_kshapid=/bin/tee
		#typeset tet_l14_kshapid=\${TETKSHAPID}
		typeset -i tet_l14_pid=0

		if test ! -x \$tet_l14_kshapid
		then
			print "\$tet_l14_kshapid is not executable"
			exit 1
		fi

		\$tet_l14_kshapid \${1:+"\$@"} >/dev/stderr |&
		tet_l14_pid=\$!

		sleep 1
		if kill -0 \$tet_l14_pid 2>/dev/null
		then
			TET_KSHAPID_USABLE=yes
			TET_KSHAPID_REQUEST=
			TET_KSHAPID_PID=\$tet_l14_pid
		else
			TET_KSHAPID_USABLE=no
			print "startup error in ksh API helper process"
			exit 1
		fi
		return 0
	}

	function tet_kshapid_write
	{
		[[ ! -z \${TET_DEBUG} ]] && set -x
		if print -p -R "\$*"
		then
			: ok
		else
			TET_KSHAPID_USABLE=no
			print "write error to ksh API helper process" \
				"the line that cannot be written is: \"\$*\""
		fi
	}

	function tet_kshapid_request
	{
		tet_kshapid_write tet_request 3.8 \$*
	}

	function tet_kshapid_tpstart
	{
		tet_kshapid_request tet_tpstart \${1:?}
		tet_kshapid_release
		return 0
	}

	function tet_kshapid_release
	{
		return 0
	}

	function tet_api_init
	{
		[[ ! -z \${TET_DEBUG} ]] && set -x
		print OUTFILE = \$OUTFILE
		tet_kshapid_startup \${1:+"\$@"}
	}

	function tet_tpstart
	{
		[[ ! -z \${TET_DEBUG} ]] && set -x
		typeset -i tet_l3_testnum=\${1:?}

		if tet_kshapid_tpstart \$tet_l3_testnum
		then
			: ok
		else
			print "tet_kshapid_tpstart failed for \$tet_l3_testnum"
			return 1
		fi

		return 0
	}

	OUTFILE=\$mydir/\${me%%.ksh*}.out
	TETKSHAPID=\$mydir/ctikshapid

	tet_api_init

	# NUMBER_OF_TESTS and TET_TESTNAMES are initialized by the caller
	typeset -i testnum=0
	while ((testnum < NUMBER_OF_TESTS)); do
		[[ ! -z \${TET_DEBUG} ]] && set -x
		tet_thistest=\${TET_TESTNAMES[\$testnum]}

		tet_tpstart \$testnum
		TET_TPSTART_RC=\$?

		if test \$TET_TPSTART_RC -eq 0
		then
			(
				[[ ! -z \${TET_DEBUG} ]] && set -x

				"\$tet_thistest"
			)
		else
			print tp \$testnum has been deleted
		fi
		((testnum = testnum + 1))
	done
EOF

	${SHCOMP}    'test1script.sh' 'test1script.shbin'
	${SHCOMP} -n 'test1script.sh' 'test1script_n.shbin'
	
	for testscript in \
		'test1script.sh' \
		'test1script.shbin' \
		'test1script_n.shbin' ; do
		out.stderr="${ { out.stdout="${ ${SHELL} "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		(( out.res == 0 )) || err_exit "test1: Test shell returned with exit code ${out.res}"
		[[ "${out.stdout}" == OUTFILE\ =*          ]] || err_exit "test1: stdout does not match OUTFILE\ =*, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" != ''                   ]] || err_exit 'test1: stderr empty'
		[[ "${out.stderr}" != *redirection*failed* ]] || err_exit 'test1: stderr contains *redirection*failed*'
		[[ "${out.stderr}" != *no*query*process*   ]] || err_exit 'test1: stderr contains *no*query*process*'
		[[ "${out.stderr}" == *IN*CONFIG*ROUTINE*  ]] || err_exit 'test1: stderr does not match *IN*CONFIG*ROUTINE*'
		[[ "${out.stderr}" == *IN*RUN*ROUTINE*     ]] || err_exit 'test1: stderr does not match *IN*RUN*ROUTINE*'
	done

	rm 'test1script.sh' 'test1script.shbin' 'test1script_n.shbin'
	
	return 0
}

# simple test for error messages
function test2
{
	compound out=( typeset stdout stderr ; integer res )
	typeset testscript

cat >'mypipe.sh' <<EOF
	function mypipe
	{
	       sleep 15
	}

	# launch child
	mypipe |&

	print -p 'hello'

	# this command substitution causes ast-ksh.2010-03-09 to loose
	# the pipe to the coprocess
	z="\$( /bin/true \$(/bin/true) )"

	print -p 'world'

	# done, terminate processing child
	kill %1
	wait
	
	print 'done'

	exit 0
EOF

	${SHCOMP}    'mypipe.sh' 'mypipe.shbin'
	${SHCOMP} -n 'mypipe.sh' 'mypipe_n.shbin'

	for testscript in \
		'mypipe.sh' \
		'mypipe.shbin' \
		'mypipe_n.shbin' ; do
		out.stderr="${ { out.stdout="${ ${SHELL} "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		(( out.res == 0 )) || err_exit "test2: Test shell returned with exit code ${out.res}"
		[[ "${out.stdout}" == 'done' ]] || err_exit "test2: Expected stdout match 'done', got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == ''     ]] || err_exit "test2: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
	done

	rm 'mypipe.sh' 'mypipe.shbin' 'mypipe_n.shbin'

	return 0
}


# simple test with coprocess duing useful stuff
function test3
{
	compound out=( typeset stdout stderr ; integer res )
	typeset testscript

cat >'sumpipe.sh' <<EOF
	function sumpipe
	{
	      compound var

	      # read objects from pipe
	      while read -C var ; do
        	      float var.sum

        	      # calculate sum
        	      (( var.sum=var.a+var.b ))

        	      # send object back
        	      print -C var
	      done
	}

	# launch child
	sumpipe |&

	# create object
	compound x=(
	      float a
	      float b
	)

	(( x.a=1.1 , x.b=2.2 ))

	# send objects for processing
	print -p -C x

	# this command substitution causes ast-ksh.2010-03-09 to loose
	# the pipe to the coprocess
	z="\$(print -- \$(date) \$(hostname) )"

	(( x.a=4.4 ))
	print -p -C x


	# receive answer and print it human readable -v (-C is compressed format)
	read -C -p y ; print -v y
	read -C -p y ; print -v y

	# done, terminate processing child
	kill %1
	wait
EOF

	${SHCOMP}    'sumpipe.sh' 'sumpipe.shbin'
	${SHCOMP} -n 'sumpipe.sh' 'sumpipe_n.shbin'

	for testscript in \
		'sumpipe.sh' \
		'sumpipe.shbin' \
		'sumpipe_n.shbin' ; do
		out.stderr="${ { out.stdout="${ ${SHELL} "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		(( out.res == 0 )) || err_exit "test3: Test shell returned with exit code ${out.res}"
		[[ "${out.stdout}" == *sum=3.3*sum=6.6* ]] || err_exit "test3: Expected stdout match *sum=3.3*sum=6.6*, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == ''                ]] || err_exit "test3: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
	done

	rm 'sumpipe.sh' 'sumpipe.shbin' 'sumpipe_n.shbin'

	return 0
}


# run tests
test1
test2
test3


# cleanup
cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."


# tests done
exit $((Errors<125?Errors:125))
