#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This module tests whether making IFS a static variable works
#
# This was reported as CR #6990612 ("Unset local IFS variable crash"):
# ------------ snip ------------
# Description
#    the testcase
# shows this:
#  $ ksh93 ~/crash.ksh
#  Core dumped
# Frequency
#     Always
# Regression
#     no
# Steps to Reproduce
#     Run testcase:
# #!/bin/ksh93
# print ">$IFS<"
# function crash
# {
#   typeset L_FILE
#   typeset L_VALIDATION
#   typeset L_VARIABLE
#   typeset L_MOD IFS
#   OS=$(uname)
# }
# crash
# print ">$IFS<"
# function crash2
# {
#   typeset IFS
#   true
#   unset IFS
#   echo a b c | while read x ; do
#     echo $x
#   done
# }
# crash2
# print ">$IFS<"
# Expected Result
#     No crash
# Actual Result
#     Crash
# Error Message(s)
#     Core dumped
# Test Case
#     #!/bin/ksh93
# print ">$IFS<"
# function crash
# {
#   typeset L_FILE
#   typeset L_VALIDATION
#   typeset L_VARIABLE
#   typeset L_MOD IFS
#   OS=$(uname)
# }
# crash
# print ">$IFS<"
# function crash2
# {
#   typeset IFS
#   true
#   unset IFS
#   echo a b c | while read x ; do
#     echo $x
#   done
# }
# crash2
# print ">$IFS<"
# Workaround
#     
# Additional configuration information
#     Solaris 11, tosiba laptop
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_sun_solaris_cr_6990612_unset_local_ifs_variable_crash.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function test1
{
	cat >'testscript1.sh' <<EOF
echo ">\$IFS<"

function crash
{
 typeset L_FILE
 typeset L_VALIDATION
 typeset L_VARIABLE
 typeset L_MOD IFS

 OS=\$(uname)
}

crash

echo ">\$IFS<"

function crash2
{
 typeset IFS
 true
 unset IFS
 echo a b c | while read x ; do
   echo \$x
 done
}

crash2

echo ">\$IFS<"
EOF

	typeset -r expected_output=$'> \t\n<\n> \t\n<\na b c\n> \t\n<'
	compound out=( typeset stdout stderr ; integer res )
	typeset testname="${0}"
	typeset s=''
	integer i
	integer saved_Errors
	typeset testscript
	
	# provide a compiled version of the script
	${SHCOMP}    'testscript1.sh' 'testscript1.shbin'
	${SHCOMP} -n 'testscript1.sh' 'testscript1_n.shbin'
	
	for testscript in \
		'testscript1.sh' \
		'testscript1.shbin' \
		'testscript1_n.shbin' ; do
		(( saved_Errors=Errors ))

		# The crash depends on a precise memory layout and allocation/deallocation
		# sequence. To ensure we trigger the error we pass the shell an environment
		# variable called "s" with a ever-growing (per iteration) string length
		for (( i=2 ; i < 1024 && (Errors-saved_Errors) < 2 ; i++ )) ; do
			out.stderr="${ { out.stdout="${ s="${s}${i}" ; s="$s" ${SHELL} -o nounset "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

			[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
			[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
			(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
		done
	done

	rm 'testscript1.sh' 'testscript1.shbin' 'testscript1_n.shbin'

	return 0
}


# run tests
test1


cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
