#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# Test module to verify that function and variable names are
# CSI (Code Set Independent)-enabled, e.g. that they can handle
# characters outside ASCII for such names
#


# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_script_csi.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }



# basic functional tests, with en_US.UTF-8 locale
function test_basic1
{
	typeset testname
	compound out=( typeset stdout stderr ; integer res )
	compound -r -a tests=(
		# POSIX function with Chinese name
		(
			cmd_fmt='\u[5929]\u[6cb3]_\u[53f7]() { print "OK" ; } ; \u[5929]\u[6cb3]_\u[53f7]\n'
			expected_output='OK'
		)
		# ksh function with Chinese name
		(
			cmd_fmt='function \u[5929]\u[6cb3]_\u[53f7] { print "OK" ; } ; \u[5929]\u[6cb3]_\u[53f7]\n'
			expected_output='OK'
		)
		# plain variable with Chinese name
		(
			cmd_fmt='\u[5929]\u[6cb3]_\u[53f7]="OK" ; print "${\u[5929]\u[6cb3]_\u[53f7]}"\n'
			expected_output='OK'
		)
		# compound variable with Chinese name
		(
			cmd_fmt='compound \u[5929]\u[6cb3] ; \u[5929]\u[6cb3].\u[53f7]="OK" ; print "${\u[5929]\u[6cb3].\u[53f7]}"\n'
			expected_output='OK'
		)
		# associative array variable with Chinese name
		(
			cmd_fmt='typeset -A \u[5929]\u[6cb3] ; \u[5929]\u[6cb3][\u[53f7]]="OK" ; print "${\u[5929]\u[6cb3][\u[53f7]]}"\n'
			expected_output='OK'
		)
	)
	integer ti # test index
	
	for (( ti=0 ; ti < ${#tests[@]} ; ti++ )) ; do
		nameref test=tests[ti]
		
		# plain test
		testname="${0}/${ti}/plain"
		out.stderr="${ { out.stdout="${ printf "${test.cmd_fmt}" | ${SHELL} ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${test.expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${test.expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ;}"
		[[ "${out.stderr}" == ''                        ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ;}"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"

		# same as above, but uses "shcomp"
		testname="${0}/${ti}/shcomp"
		out.stderr="${ { out.stdout="${ printf "${test.cmd_fmt}" | ${SHCOMP} | ${SHELL} ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${test.expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${test.expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ;}"
		[[ "${out.stderr}" == ''                        ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ;}"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"

		# same as above, but uses "shcomp -n"
		testname="${0}/${ti}/shcomp_n"
		out.stderr="${ { out.stdout="${ printf "${test.cmd_fmt}" | ${SHCOMP} -n | ${SHELL} ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${test.expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${test.expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ;}"
		[[ "${out.stderr}" == ''			]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ;}"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}

# setup
builtin cp
builtin mkdir
builtin rm
builtin rmdir

# test whether the locale uses an UTF-8 (-like) encoding and override it on demand
typeset -r utf8_euro_char1=$'\u[20ac]'
typeset -r utf8_euro_char2=$'\342\202\254'
(( (${#utf8_euro_char1} == 1) && (${#utf8_euro_char2} == 1) )) || export LC_ALL='en_US.UTF-8'
if [[ "$(printf '\u[20ac]')" != $'\342\202\254' ]] ; then
        err_exit 'Locale overrride failed.'
	cd "${ocwd}"
	rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."
        exit $((Errors<125?Errors:125))
fi

# run tests
test_basic1


cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
