#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# Test module for ksh93/bash localised string literals ($"...")
#


# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_script_l10n.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }


typeset -T test_t=(
	typeset testname
	typeset -a cmd_args
	typeset expected_stdout
	typeset expected_stderr
	integer expected_res
	
	function run_test
	{
		compound out=( typeset stdout stderr ; integer res )
		
		out.stderr="${ { out.stdout="${ "${_.cmd_args[@]}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		[[ "${out.stdout}" == ${_.expected_stdout} ]] || err_exit "${_.testname}: Expected stdout==${ printf '%q\n' "${_.expected_stdout}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == ${_.expected_stderr} ]] || err_exit "${_.testname}: Expected stderr==${ printf '%q\n' "${_.expected_stderr}" ;}, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == _.expected_res )) || err_exit "${_.testname}: Unexpected exit code ${out.res}"
		return 0
	}
)

# basic functional tests, with french locale
function test_basic1
{
	compound cleanobjs=(
		typeset -a files
		typeset -a dirs
	)

	test_t -a tests=(
		(
			testname="${0}:t1"
			cmd_args=( $SHELL -c 'unset ${!LC_*} ; export LC_MESSAGES="fr_FR.UTF-8" ; $SHELL "${testscript}"' )
			expected_stdout=$'Simple demonstration of ksh93 message translation\nMessage locale is: fr_FR.UTF-8\nBonjour\nAu Revoir\nBienvenu David\nThis string is not translated because it is not in the message catalog\n#####\nSimple demonstration of ksh93 message translation\nMessage locale is: C\nHello\nGoodbye\nWelcome David\nThis string is not translated because it is not in the message catalog'
		)
		(
			testname="${0}:t1_unuseable_external_nlspath"
			cmd_args=( $SHELL -c 'unset ${!LC_*} ; export LC_MESSAGES="fr_FR.UTF-8" NLSPATH="/tmp/dummy/%N.cat" ; $SHELL "${testscript}"' )
			expected_stdout=$'Simple demonstration of ksh93 message translation\nMessage locale is: fr_FR.UTF-8\nBonjour\nAu Revoir\nBienvenu David\nThis string is not translated because it is not in the message catalog\n#####\nSimple demonstration of ksh93 message translation\nMessage locale is: C\nHello\nGoodbye\nWelcome David\nThis string is not translated because it is not in the message catalog'
		)
	)
	
	integer i

	# write test script
	cleanobjs.files+=( 'msg.sh' )
	cat >'msg.sh' <<EOF
export NLSPATH=\$PWD/locale/%l/%N.cat

name="David"

function mymsg
{
	print  "Simple demonstration of ksh93 message translation"
	print  "Message locale is: \${LC_ALL:-\${LC_MESSAGES:-\${LANG:-C}}}"

	echo   \$"Hello"
	print  \$"Goodbye"
	printf \$"Welcome %s\n" \$name

	print  \$"This string is not translated because it is not in the message catalog"
	return 0
}

mymsg
print '#####'
export LC_ALL='C'
mymsg
EOF
	mkdir -p 'compiled' 'compiled_n' ; cleanobjs.dirs+=( 'compiled' 'compiled_n' )
	$SHCOMP    'msg.sh' 'compiled/msg.sh'   ; cleanobjs.files+=( 'compiled/msg.sh' )
	$SHCOMP -n 'msg.sh' 'compiled_n/msg.sh' ; cleanobjs.files+=( 'compiled_n/msg.sh' )

	cleanobjs.files+=( 'msg.sh.C.msg' )
	printf '$quote "\n$set 3\n1 "Hello"\n2 "Goodbye"\n3 "Welcome %%s\\\\n"\n' >'msg.sh.C.msg'

	cleanobjs.files+=( 'msg.sh.fr.msg' )
	printf '$quote "\n$set 3\n1 "Bonjour"\n2 "Au Revoir"\n3 "Bienvenu %%s\\\\n"\n' >'msg.sh.fr.msg'

	mkdir -p 'locale/C' 'locale/fr' ; cleanobjs.dirs+=( 'locale/C' 'locale/fr' 'locale' )
	msggen 'locale/C/mymsg.cat'  'msg.sh.C.msg'  ; cleanobjs.files+=( 'locale/C/mymsg.cat' )
	msggen 'locale/fr/mymsg.cat' 'msg.sh.fr.msg' ; cleanobjs.files+=( 'locale/fr/mymsg.cat' )
	
	export testscript=''
	
	for testscript in \
		'msg.sh' \
		'compiled/msg.sh' \
		'compiled_n/msg.sh' ; do
		for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
			tests[i].run_test
		done
	done

	rm -- "${cleanobjs.files[@]}"
	rmdir -- "${cleanobjs.dirs[@]}"
	return 0
}

function test_per_function_catalogs
{
	compound cleanobjs=(
		typeset -a files
		typeset -a dirs
	)
	integer i
	
	test_t -a tests=(
		(
			# plain test with an invalid NLSPATH
			testname="${0}:t1"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				'/dummy/%N.cat'
				''
				''
				)
			expected_stdout=$'This is function XXX\nThis is function XXX'
		)
		(
			# test per-function catalogs. The POSIX shell standard says that 
			# shell functions should act like independent external commands, therefore
			# ksh93 allows that each functioncan have it's NLSPATH, locale and resulting catalog
			# (locale/%l/%N.cat means that each shell function name is used in %N)
			testname="${0}:t1_per_function_catalog"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				"$PWD/locale/%l/%N.cat"
				''
				''
				)
			expected_stdout=$'This is function f_a\nThis is function f_b'
		)
		(
			# same as previous test, but only set NSLPATH for function "f_a"
			# (locale/%l/%N.cat means that each shell function name is used in %N)
			testname="${0}:t1_per_function_catalog_only_for_f_a"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				''
				"$PWD/locale/%l/%N.cat"
				''
				)
			expected_stdout=$'This is function f_a\nThis is function XXX'
		)
		(
			# same as previous test, but only set NSLPATH for function "f_b"
			# (locale/%l/%N.cat means that each shell function name is used in %N)
			testname="${0}:t1_per_function_catalog_only_for_f_b"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				''
				''
				"$PWD/locale/%l/%N.cat"
				)
			expected_stdout=$'This is function XXX\nThis is function f_b'
		)
		(
			# test a single catalog for the whole script
			testname="${0}:t1_single_catalog_per_script"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				"$PWD/locale/%l/msg.sh.cat"
				''
				''
				)
			expected_stdout=$'This is a function in msg.sh\nThis is a function in msg.sh'
		)
		(
			# test a single catalog for the whole script but override NSLPATH for function "f_a"
			testname="${0}:t1_single_catalog_per_script"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				"$PWD/locale/%l/msg.sh.cat"
				"$PWD/locale/%l/%N.cat"
				''
				)
			expected_stdout=$'This is function f_a\nThis is a function in msg.sh'
		)
		(
			# test a single catalog for the whole script but override NSLPATH for function "f_b"
			testname="${0}:t1_single_catalog_per_script"
			cmd_args=(
				$SHELL
				-c
				'unset ${!LC_*} LANG NLSPATH ; export LC_MESSAGES="fr_FR.UTF-8" test_NLSPATH="$1" test_NLSPATH_f_a="$2" test_NLSPATH_f_b="$3" ; $SHELL "${testscript}"'
				''
				"$PWD/locale/%l/msg.sh.cat"
				''
				"$PWD/locale/%l/%N.cat"
				)
			expected_stdout=$'This is a function in msg.sh\nThis is function f_b'
		)
	)

	# write test script
	cleanobjs.files+=( 'msg.sh' )
	cat >'msg.sh' <<EOF
export NLSPATH="\${test_NLSPATH}"

function f_a
{
	printf \$"This is function XXX\n"
	return 0
}

function f_b
{
	printf \$"This is function XXX\n"
	return 0
}

if [[ "\${test_NLSPATH_f_a}" != '' ]] ; then
	NLSPATH="\${test_NLSPATH_f_a}" f_a
else
	f_a
fi
if [[ "\${test_NLSPATH_f_b}" != '' ]] ; then
	NLSPATH="\${test_NLSPATH_f_b}" f_b
else
	f_b
fi
EOF
	# provide a compiled version of the script
	mkdir -p 'compiled' 'compiled_n' ; cleanobjs.dirs+=( 'compiled' 'compiled_n' )
	$SHCOMP   'msg.sh'  'compiled/msg.sh'   ; cleanobjs.files+=( 'compiled/msg.sh' )
	$SHCOMP -n 'msg.sh' 'compiled_n/msg.sh' ; cleanobjs.files+=( 'compiled_n/msg.sh' )

	# build catalogs
	mkdir -p 'locale/C' 'locale/fr' ; cleanobjs.dirs+=( 'locale/C' 'locale/fr' 'locale' )

	cleanobjs.files+=( 'msg.sh.C.msg' )
	printf '$quote "\n$set 3\n1 "This is function XXX\\\\n"\n2 "This is function XXX\\\\n"\n' >'msg.sh.C.msg'
	msggen 'locale/C/msg.sh.cat'  'msg.sh.C.msg'  ; cleanobjs.files+=( 'locale/C/msg.sh.cat' )
	cp 'locale/C/msg.sh.cat' 'locale/C/f_a.cat'   ; cleanobjs.files+=( 'locale/C/f_a.cat' )
	cp 'locale/C/msg.sh.cat' 'locale/C/f_b.cat'   ; cleanobjs.files+=( 'locale/C/f_b.cat' )

	cleanobjs.files+=( 'msg.sh.fr.msg' )
	printf '$quote "\n$set 3\n1 "This is a function in msg.sh\\\\n"\n2 "This is a function in msg.sh\\\\n"\n' >'msg.sh.fr.msg'
	msggen 'locale/fr/msg.sh.cat' 'msg.sh.fr.msg' ; cleanobjs.files+=( 'locale/fr/msg.sh.cat' )

	cleanobjs.files+=( 'f_a.fr.msg' )
	printf '$quote "\n$set 3\n1 "This is function f_a\\\\n"\n2 "This is function f_a\\\\n"\n' >'f_a.fr.msg'
	msggen 'locale/fr/f_a.cat' 'f_a.fr.msg' ; cleanobjs.files+=( 'locale/fr/f_a.cat' )

	cleanobjs.files+=( 'f_b.fr.msg' )
	printf '$quote "\n$set 3\n1 "This is function f_b\\\\n"\n2 "This is function f_b\\\\n"\n' >'f_b.fr.msg'
	msggen 'locale/fr/f_b.cat' 'f_b.fr.msg' ; cleanobjs.files+=( 'locale/fr/f_b.cat' )
	
	export testscript=''
	for testscript in \
		'msg.sh' \
		'compiled/msg.sh' \
		'compiled_n/msg.sh' ; do
		for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
			tests[i].run_test
		done
	done

	rm -- "${cleanobjs.files[@]}"
	rmdir -- "${cleanobjs.dirs[@]}"
	return 0
}


function test_l10nliteral_to_variable_assignment
{
	compound cleanobjs=(
		typeset -a files
		typeset -a dirs
	)
	compound out=( typeset stdout stderr ; integer res )
	typeset testscript
	typeset testname

	# write test script
	cleanobjs.files+=( 'msg.sh' )
	cat >'msg.sh' <<EOF
unset \${!LC_*} LANG
export LC_MESSAGES='fr_FR.UTF-8'
export NLSPATH="\${PWD}/locale/%l/msg.sh.cat"
integer res=0

# plain function with localised literal
function f_a
{
	printf \$"This is a string\n"
	return 0
}

# same function as "f_a" but stores the localised string literal in
# a local variable ("fmt") first before using it
function f_b
{
	typeset fmt=\$"This is a string\n"
	printf "\${fmt}"
	return 0
}

f_a >'f_a.out'
f_b >'f_b.out'

if ! cmp -s 'f_a.out' 'f_b.out' ; then
	printf 'Files f_a.out and f_b.out differ, %q vs. %q\n' "\$( < 'f_a.out' )" "\$( < 'f_b.out' )"
	(( res=1 ))
else
	printf '#match\n'
fi

rm 'f_a.out' 'f_b.out'

exit \${res}
EOF

	# provide a compiled version of the script
	mkdir -p 'compiled' 'compiled_n' ; cleanobjs.dirs+=( 'compiled' 'compiled_n' )
	$SHCOMP    'msg.sh' 'compiled/msg.sh'   ; cleanobjs.files+=( 'compiled/msg.sh' )
	$SHCOMP -n 'msg.sh' 'compiled_n/msg.sh' ; cleanobjs.files+=( 'compiled_n/msg.sh' )

	# build catalogs
	mkdir -p 'locale/C' 'locale/fr' ; cleanobjs.dirs+=( 'locale/C' 'locale/fr' 'locale' )

	cleanobjs.files+=( 'msg.sh.C.msg' )
	printf '$quote "\n$set 3\n1 "This is a string\\\\n"\n2 "This is a string\\\\n"\n' >'msg.sh.C.msg'
	msggen 'locale/C/msg.sh.cat'  'msg.sh.C.msg'  ; cleanobjs.files+=( 'locale/C/msg.sh.cat' )

	cleanobjs.files+=( 'msg.sh.fr.msg' )
	printf '$quote "\n$set 3\n1 "This is a string in msg.sh\\\\n"\n2 "This is a string in msg.sh\\\\n"\n' >'msg.sh.fr.msg'
	msggen 'locale/fr/msg.sh.cat' 'msg.sh.fr.msg' ; cleanobjs.files+=( 'locale/fr/msg.sh.cat' )
	
	for testscript in \
		'msg.sh' \
		'compiled/msg.sh' \
		'compiled_n/msg.sh' ; do
		testname="${0}/${testscript}"
		out.stderr="${ { out.stdout="${ $SHELL "$testscript" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		# printf '%s\n' "${out.stdout}" is intentional here
		[[ "${out.stdout}" == '#match'  ]] || err_exit "${testname}: Expected stdout=='#match', got: ${ printf '%s\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == ''        ]] || err_exit "${testname}: Expected empty stderr, got: ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done

	rm -- "${cleanobjs.files[@]}"
	rmdir -- "${cleanobjs.dirs[@]}"
	return 0
}

# setup
builtin cp
builtin mkdir
builtin rm
builtin rmdir

# Add PATH element if we run this module on an ast-open build...
tmppath="$(set +o noglob ; printf '%s\n' ${ocwd}/arch/*.*/bin)"
if [[ -d "${tmppath}" ]] ; then
	PATH+=":${tmppath}"
else
	# ... otherwise we try to find the path to "msggen"&co.
	# by other means
	case "$(uname -s)" in
		'SunOS')
			PATH+=':/usr/ast/bin' # for Solaris >= 11
			;;
	esac
fi

which 'msggen' >'/dev/null' || err_exit 'msggen not found in PATH'
[[ "$(locale -a)" == ~(E)(fr_FR.UTF-8|fr_FR.utf8) ]] || err_exit 'Locale fr_FR.UTF-8 not found.'

# stop here if we have any errors at this point
if (( Errors > 0 )) ; then
	cd "${ocwd}"
	rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."
	exit $((Errors<125?Errors:125))
fi

# run tests
test_basic1
test_per_function_catalogs
test_l10nliteral_to_variable_assignment

cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
